import GObject from 'gi://GObject';
import St from 'gi://St';
import Clutter from 'gi://Clutter';
import Pango from 'gi://Pango';
import Glib from 'gi://GLib';

import * as Main from 'resource:///org/gnome/shell/ui/main.js';
import {Extension} from 'resource:///org/gnome/shell/extensions/extension.js';
import * as Util from 'resource:///org/gnome/shell/misc/util.js';
import * as animationUtils from 'resource:///org/gnome/shell/misc/animationUtils.js';

import * as Utils from './Utils.js';

import {OpenAiClient} from './OpenAiClient.js';
const openaiClient = new OpenAiClient();

const OVERLAY_HEIGHT = 600;
const OVERLAY_WIDTH = 300;
const QUESTION_TEXT_MAX_LENGTH = 10000;
var SCROLL_TIME = 100;

const ICON_LOADING_FINISHED = "object-select-symbolic";
const ICON_LOADING = "content-loading-symbolic"

const ChatMessage = GObject.registerClass(
class ChatMessage extends St.Bin {
    _init(params) {
        super._init({
            style_class: 'message'
        });

        this._text = new Clutter.Text({
            text: params.text,
            single_line_mode: false,
            line_wrap: true,
            line_wrap_mode: Pango.WrapMode.WORD,
            editable: false,
            selectable: true,
            margin_top: 10,
            margin_left: 10,
            font_name: "Sans 10",
            margin_right: 10,
            margin_bottom: 10,
            cursor_visible: true,
            use_markup: true,
            selection_color: new Clutter.Color({red: 255, green: 153, blue: 51, alpha: 100}),
            reactive: true
        })

        this._text.connect('key-press-event', (object, event) => {
            let code = event.get_key_code();
            let state = event.get_state();

            // Ctrl + A
            if (state === 4 && code === 38) {
                object.set_selection(0, object.text.length);
                return true;
            }
            // Ctrl + C
            else if (state === 4 && code === 54) {
                let clipboard = St.Clipboard.get_default();
                let selection = object.get_selection();
                clipboard.set_text(St.ClipboardType.CLIPBOARD, selection);
                return true;
            }
        });

        this.add_actor(this._text);
    }

    vfunc_paint(paint_context) {
        super.vfunc_paint(paint_context);

        if (Utils.isGnomeDarkModeEnabled()) {
            this._text.color = new Clutter.Color({red: 200, green: 200, blue: 200, alpha: 255})
        } else {
            this._text.color = new Clutter.Color({red: 20, green: 20, blue: 20, alpha: 255})
        }
    }
});


export class Overlay {

    constructor() {
        this.questionHistory = [];
        this.questionHistoryIndex = 0;
        this.isLoading = false;
	this.extensionObject = Extension.lookupByUUID('openai-gnome@etixsoftware.de');

        this.overlay = new St.BoxLayout({
            style_class: 'modal-dialog',
            vertical: true,
            reactive: true
        });

        this.overlay.connect('key-press-event', (object, event) => {
            let code = event.get_key_code();
            let state = event.get_state();

            // close overlay if ESC is pressed
            if (state === 0 && code === 9) {
                this.extensionObject._onTogglePress();
                return true;
            }

            return false;
        });

	Main.layoutManager.connectObject('monitors-changed',
            this._updateLayout.bind(this));
	
        this.overlay.visible = false;
	this.overlay.connect('notify::allocation',
            Main.layoutManager._queueUpdateRegions.bind(Main.layoutManager));

        Main.layoutManager.addChrome(this.overlay, {affectsInputRegion: false, affectsStruts: true, trackFullscreen: true });

        this._updateLayout();
        this._buildChatRow();
	this._buildQuestionRow();
	this._buildMenuRow();

        let showWelcomeBanner = Utils.getSettings().get_boolean("show-welcome-banner");
        
        this._appendChatMessage("Welcome to Bianbu Chat.");
        let token = Utils.getSettings().get_string("openai-api-key");
        if (token.length === 0)
            this._appendChatMessage("You need a API-Key to use this extension. Please open settings dialog (or type ':settings') and paste your API-Key.");
        
	this._appendChatMessage("Shortcuts:\n" +
            "- CTRL + L\tClear History\n" +
            "- ESC\t\tHide Overlay ")

    }

    _buildQuestionRow() {
        this.questionRow = new St.BoxLayout({
            style_class: 'input-container',
            vertical: false,
            x_expand: true
        });

        this.inputQuestion = new St.MultilineEntry({
            style_class: 'question-input',
            can_focus: true,
            hint_text: 'Ask Bianbu Chat...',
            track_hover: true,
            x_expand: true
        });
	
        this.inputQuestion.clutter_text.connect('key-press-event', (actor, event)  => {
            if (Object.getPrototypeOf(Main.inputMethod.currentFocus) !== this.inputMethod) {
                this._enterQuestion();
            }
        });

	this.inputQuestion.connect('button-press-event', this._enterQuestion.bind(this));
        this.inputQuestion.connect('key-press-event', (object, event) => {
            let code = event.get_key_code();
            let state = event.get_state();
            // Ctrl + L
            if ((state && 4 === 4) && code === 46) {
                this._onClearChatHistoryPress();
                return true;
            }

            // UP
            if (state === 0 && code === 111) {
                this._chatHistoryScroll(1);
                return true;
            }

            // DOWN
            if (state === 0 && code === 116) {
                this._chatHistoryScroll(-1);
                return true;
            }

	    // Ctrl + Enter
            if ((state && 4 === 4) && code === 36) {
                this._onQuestionEnter();
                return true;
            }

        });

        this.questionRow.add(this.inputQuestion);
        this.inputQuestion.clutter_text.connect('activate', this._onQuestionEnter.bind(this));
	this.inputQuestion.clutter_text.connect('text-changed', this._onTextChanged.bind(this));
	this.inputQuestion.clutter_text.connect('cursor-changed', this._questionCursorTrackScrollbar.bind(this));
	this.inputQuestion.clutter_text.set_max_length(QUESTION_TEXT_MAX_LENGTH);
        this.questionScrollView = new St.ScrollView({
            style_class: 'vfade',
            enable_mouse_scrolling: true,
            vscrollbar_policy: St.PolicyType.ALWAYS,
            hscrollbar_policy: St.PolicyType.NEVER,
            overlay_scrollbars: false,
            y_expand: false,
            x_expand: true
        });
        this.questionScrollView.set_height(150);
        this.overlay.add_child(this.questionScrollView);


        this.questionScrollView.add_actor(this.questionRow);
    }

    _buildChatRow() {
        this.chatRow = new St.ScrollView({
            style_class: 'vfade',
            enable_mouse_scrolling: true,
            vscrollbar_policy: St.PolicyType.ALWAYS,
            hscrollbar_policy: St.PolicyType.NEVER,
            overlay_scrollbars: false,
            y_expand: true,
            x_expand: true
        });

        this.overlay.add_child(this.chatRow);

        this.chatContainer = new St.BoxLayout({
            vertical: true,
            x_expand: true,
            y_expand: false,
        });

        this.chatRow.add_actor(this.chatContainer);
    }

    _buildMenuRow() {
        this.menuRow = new St.BoxLayout({
            style_class: 'controls-container',
            vertical: false,
	    x_expand: true
        });

        this.btnClearHistory = new St.Button({
            label: "Clear History",
            style_class: 'button',
        });

        this.btnClearHistory.connect('button-press-event', this._onClearChatHistoryPress.bind(this));

        this.menuRow.add_child(this.btnClearHistory);

        this.btnSpace = new St.Button({
             label: "",
	     x_expand: true
         });
	this.menuRow.add_child(this.btnSpace);

        this.btnSendQuestion = new St.Button({
             label: "Send",
             style_class: 'button',
         });
         this.btnSendQuestion.connect('button-press-event', this._onQuestionEnter.bind(this));

        this.menuRow.add(this.btnSendQuestion);

        this.overlay.add_child(this.menuRow);
    }

    _onQuestionEnter() {
        let question = this.inputQuestion.text;

        if (question.length === 0 || this.isLoading) {
            return;
        }

        this.isLoading = true;

        this.questionHistory.push(question);
        this.questionHistoryIndex = 0;

        if (question === ':settings') {
            this.inputQuestion.set_text("");
            this.extensionObject._onTogglePress();
            Extension.openPreferences();
            return;
        }

        openaiClient.ask(question).then(answer => {
            let chat = "<b>You:</b> " + question.trim() + "\n\n" + "<b>BianbuChat:</b> " + answer.trim();

            this.inputQuestion.set_text("");

            this._appendChatMessage(chat);

        }).catch(error => {
            console.error(error);
            this._appendChatMessage("Error while try to access OpenAI: " + error);
        }).finally(() => {
            this.isLoading = false;
        })
    }

    _appendChatMessage(msg) {
        let chatEntry = new ChatMessage({
            text: msg
        });

        let chatEntryContainer = new St.Bin({
            style_class: "chat-entry-container"
        });
        chatEntryContainer.add_actor(chatEntry);

        this.chatContainer.add(chatEntryContainer);

        animationUtils.ensureActorVisibleInScrollView(this.chatRow, chatEntry);
    }

    _onClearChatHistoryPress() {
        this.inputQuestion.set_text("");
        openaiClient.clearHistory();
        let chats = this.chatContainer.get_children();

        chats.forEach(c => this.chatContainer.remove_child(c));
	this._appendChatMessage("Welcome to Bianbu Chat.");
    }

    _chatHistoryScroll(delta) {
        if ((this.questionHistoryIndex + delta) <= 0) {
            this.inputQuestion.set_text("");
            this.questionHistoryIndex = 0;
            return;
        }

        if ((this.questionHistoryIndex + delta) > this.questionHistory.length)
            return;

        this.questionHistoryIndex += delta;

        let index = this.questionHistory.length - this.questionHistoryIndex;

        let question = this.questionHistory[index];

        this.inputQuestion.set_text(question);
    }

    _updateLayout() {
        let monitor = Main.layoutManager.primaryMonitor;
        this.overlay.width = OVERLAY_WIDTH;
        this.overlay.height = monitor.height - 40;

        // display the overlay in the center of the screen
        this.overlay.set_position(
            monitor.x + Math.floor(monitor.width - this.overlay.width),
            monitor.y + Math.floor(monitor.height - this.overlay.height) - 10
        );
	
    }
    
    _enterQuestion() {
	global.stage.set_key_focus(this.inputQuestion);
    }

    _onTextChanged() {
	if (this.inputQuestion.clutter_text.get_text().length > QUESTION_TEXT_MAX_LENGTH - 1) {
		this.inputQuestion.clutter_text.delete_text(0,-1);
		this.inputQuestion.clutter_text.insert_text("###max length of string is 10000###", -1);
	}
    }

    _questionCursorTrackScrollbar() {
	let curPos = this.inputQuestion.clutter_text.get_cursor_position();
	let [ok, x, y, line_height] = this.inputQuestion.clutter_text.position_to_coords(curPos);
	this._ensureCursorVisibleInScrollView(this.questionScrollView, y);
    }

    _ensureCursorVisibleInScrollView(scrollView, cursorPosY) {
	let adjustment = scrollView.vscroll.adjustment;
	let [value, lower_, upper, stepIncrement_, pageIncrement_, pageSize] = adjustment.get_values();

	let offset = 0;
	let vfade = scrollView.get_effect("fade");
	if (vfade)
	    offset = vfade.fade_margins.top;

	if (cursorPosY < value + offset)
	    value = Math.max(0, cursorPosY - offset);
	else if (cursorPosY > value + pageSize - offset)
	    value = Math.min(upper, cursorPosY + offset - pageSize);
	else
	    return;

	adjustment.ease(value, {
	    mode: Clutter.AnimationMode.EASE_OUT_QUAD,
	    duration: SCROLL_TIME,
	});
    }


    show() {
	if (Main.layoutManager._findActor(this.overlay) === -1)
	    Main.layoutManager._trackActor(this.overlay, { affectsInputRegion: false, affectsStruts: true, trackFullscreen: true });
        this.overlay.visible = true;
        global.stage.set_key_focus(this.inputQuestion);
        this.inputMethod = Object.getPrototypeOf(Main.inputMethod.currentFocus);
    }

    hide() {
	Main.layoutManager._untrackActor(this.overlay);
        this.overlay.visible = false;
    }

    isVisible() {
        return this.overlay.visible;
    }

    destroy() {
        this.overlay.destroy();
    }
}
